/* This file is part of turtle, a turtle-graphics system for Scheme.

Copyright (C) 2008  Luca Saiu

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. */


#include "graphics.h"
#include "debug.h"
#include <SDL/SDL.h>
#include <SDL/SDL_thread.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdbool.h>
#include <math.h>
#include "compiletimeparameters.h"
//#include "SDL/SDL_syswm.h"

/* default_surface is mapped to the window. Only when drawing the cursor
   we copy from surface_with_cursor into default_surface, we perform an
   SDL_UpdateRect(), and then we switch back to the old default_surface: */
struct turtle_window{
  SDL_Surface *scratch_surface;
  SDL_Surface *surface_with_cursor;
  SDL_Surface *default_surface; // this is the surface mapped to the window
}; // struct

/* We always refer a window via a pointer to a malloc()ed buffer. Notice that
   we currently support only *one* window at a time: */
typedef struct turtle_window* turtle_window_t;

/* The global, one and only window we use. Note how this indirection is quite
   inefficient and not needed; I just do that so that it's easier, in the future,
   to add support for more than one window at a time: */
static turtle_window_t the_window = NULL;

/* The current expose event manager function: no function by default */
static void (*the_expose_event_handler)(void) = NULL;

void turtle_set_expose_event_handler(void (*function)(void)){
  the_expose_event_handler = function;
}

/* Manage pending events; this is currently only used for expose events: */
void turtle_manage_events_if_any(void) {
  SDL_Event event;
  
  /* Turn pending X events into SDL events: */
  SDL_PumpEvents();

  /* Do we have anything in the event queue? Keep managing events until the
     queue is empty:*/
  while(SDL_PollEvent(&event) != 0){
    switch(event.type){
    case SDL_VIDEOEXPOSE: case SDL_VIDEORESIZE: case SDL_ACTIVEEVENT:
      /* We got an expose event: call the handler function, if any: */
      if(the_expose_event_handler != NULL)
        the_expose_event_handler();
      break;
    case SDL_QUIT:
      exit(EXIT_SUCCESS);
    default:
      /* Do nothing */;
    } // switch
  } // while
  //printf("No more events\n");
}

/* Initialize SDL support, but not the graphic subsystem (yet): */
static void turtle_initialize_SDL(void){
  /* Initialize SDL, or die: */
  if(SDL_Init(0/* SDL_INIT_EVENTTHREAD */) < 0)
    turtle_fatal("Couldn't initialize SDL video support: %s\n", SDL_GetError());
  
  /* Execute SDL_Quit() when we exit (*not* now): */
  atexit(SDL_Quit);
}

/* Make a new window of the given size and return it, or fail with a 
   fatal error: */
static turtle_window_t turtle_make_window(const int width, const int height){
  /* Initialize the SDL video subsystem, thus making a window; notice that we
     can only make one window at the time in this way: */
  SDL_InitSubSystem(SDL_INIT_VIDEO);
  
  /* Set the window title: */
  SDL_WM_SetCaption("Turtle graphics", NULL /* no icon */);
  
  /* Make the window "always on top": */
  // To do: I'd like to do this, but it's not supported by SDL. I could do it
  // via X and the syswm interface, but it's ugly. Any ideas?
  
  /* Allocate space for the result: */
  turtle_window_t result = (turtle_window_t)
    turtle_malloc(sizeof(struct turtle_window));
  
  /* Make the window, and fill the structure: */
  const int preferred_depth = 0; // any depth will do
  result->default_surface =
    SDL_SetVideoMode(width, height, preferred_depth, SDL_SWSURFACE);
  if(result->default_surface == NULL)
    turtle_fatal("could not make a %ix%i window (%i bpp): %s\n",
                 width, height, preferred_depth,
                 SDL_GetError());
  
  /* Make the other surfaces we need for drawing the cursor: */
  result->surface_with_cursor =
    SDL_CreateRGBSurface(SDL_SWSURFACE, width, height,
                         result->default_surface->format->BitsPerPixel,
                         0, 0, 0, 0);
  if(result->surface_with_cursor == NULL)
    turtle_fatal("could not make the surface-with-cursor");
  result->scratch_surface =
    SDL_CreateRGBSurface(SDL_SWSURFACE, width, height,
                         result->default_surface->format->BitsPerPixel,
                         0, 0, 0, 0);
  if(result->scratch_surface == NULL)
    turtle_fatal("could not make the scratch surface");
  
  /* We're done, return our flashy new window: */
  return result;
}

int turtle_get_window_width(void){
  return the_window->surface_with_cursor->w;
}

int turtle_get_window_height(void){
  return the_window->surface_with_cursor->h;
}

void turtle_initialize_graphics(const int width, const int height){
  /* Fail if we have already initialized: */
  if(the_window != NULL)
    turtle_fatal("a window already exists");
  
  /* Initialize SDL and make the window: */
  turtle_initialize_SDL();
  the_window = turtle_make_window(width, height);
  
  /* Also clear the window, sometimes it's required (depending on the
     video mode, I think): */
  turtle_clear_window(turtle_black);
}

/* Destroy the given window, and deallocate the structure: */
static void turtle_destroy_window(const turtle_window_t window){
  /* Destroy all surfaces: */
  SDL_FreeSurface(the_window->surface_with_cursor);
  SDL_FreeSurface(the_window->default_surface);
  
  /* Finalize the SDL video subsystem, thus destroying the window: */
  SDL_QuitSubSystem(SDL_INIT_VIDEO);
  
  /* Free the buffer: */
  turtle_free(window);
}

void turtle_finalize_graphics(void){
  /* Fail if we have not finalized: */
  if(the_window == NULL)
    turtle_fatal("finalizing twice?");
  
  /* Destroy the window and finalize all SDL support: */
  turtle_destroy_window(the_window);
  the_window = NULL;
  SDL_Quit();
}

static void lock_surface_if_needed(SDL_Surface *surface){
  /* If the surface requires locking, and it's now locked, wait until
     it's freed: */
  if(SDL_MUSTLOCK(surface))
    while(SDL_LockSurface(surface) < 0)
      /* Do nothing */;
}

static void unlock_surface_if_needed(SDL_Surface *surface){
  /* If the surface requires locking (assuming that *we* are holding its
     lock) then unlock it: */
  if(SDL_MUSTLOCK(surface))
    SDL_UnlockSurface(surface);
}

/* Draw a pixel in the given depth on the given surface, without locking: */
static void turtle_draw_pixel_8bpp_unlocked(SDL_Surface *surface,
                                            int x, int y, const Uint32 color){
  Uint8 *bufp = (Uint8 *)surface->pixels + y*surface->pitch + x;
  *bufp = color;
}
static void turtle_draw_pixel_15or16bpp_unlocked(SDL_Surface *surface,
                                                 int x, int y, const Uint32 color){
  Uint16 *bufp = (Uint16 *)surface->pixels + y*surface->pitch/2 + x;
  *bufp = color;
}
static void turtle_draw_pixel_24bpp_unlocked(SDL_Surface *surface,
                                             int x, int y,
                                             Uint8 r, Uint8 g, Uint8 b){
  Uint8 *bufp =(Uint8 *)surface->pixels + y*surface->pitch + x;
  *(bufp+surface->format->Rshift/8) = r;
  *(bufp+surface->format->Gshift/8) = g;
  *(bufp+surface->format->Bshift/8) = b;
}
static void turtle_draw_pixel_32bpp_unlocked(SDL_Surface *surface,
                                             int x, int y, const Uint32 color){
  Uint32 *bufp = (Uint32 *)surface->pixels + y*surface->pitch/4 + x;
  *bufp = color;
}

void turtle_refresh(void){
  SDL_UpdateRect(the_window->default_surface, 0, 0, 0, 0);
}

void turtle_refresh_with_cursor(void){
  /* Make a copy of the surface without the cursor, which we are about
     to overwrte: */
  SDL_BlitSurface(the_window->default_surface, NULL,
                  the_window->scratch_surface, NULL);
  
  /* Copy the surface with the cursor into the default surface: */
  SDL_BlitSurface(the_window->surface_with_cursor, NULL,
                  the_window->default_surface, NULL);
  
  /* Show the default surface, which now has the cursor: */
  SDL_UpdateRect(the_window->default_surface, 0, 0, 0, 0);
  
  /* Copy the surface without the cursor that we've saved before back
     into the default surface, *and* into the surface with cursor, so
     that the next time we can simply draw another cursor without worrying
     about deleting the old one: */
  SDL_BlitSurface(the_window->scratch_surface, NULL,
                  the_window->default_surface, NULL);
  SDL_BlitSurface(the_window->scratch_surface, NULL,
                  the_window->surface_with_cursor, NULL);
}

/* Convert the given turtle-color into unsigned 8-bit components, which is the
   format used by SDL; results are written in the given locations: */
static void turtle_color_to_8bit_components(const turtle_color_t color,
                                            Uint8 *red, Uint8 *green, Uint8 *blue){
  *red = (Uint8)(color->red * 255.0);
  *green = (Uint8)(color->green * 255.0);
  *blue = (Uint8)(color->blue * 255.0);
}

/* Return true iff the pixel with the given coordinates in in the window: */
static bool is_within_bounds(const turtle_window_t window, const int x, const int y){
  return ((x >= 0) && (y >= 0) &&
          (x < window->surface_with_cursor->w) &&
          (y < window->surface_with_cursor->h));
}

static void turtle_draw_pixel_on_surface(SDL_Surface *surface,
                                         const int x, const int y,
                                         const turtle_color_t color){
  /* Convert the color into a format which SDL can use: */
  Uint8 r, g, b;
  turtle_color_to_8bit_components(color, &r, &g, &b);
  
  /* Maybe we need to lock the SDL surface: */  
  lock_surface_if_needed(surface);
  
  /* Convert the color to another internal format used by SDL, if needed: */
  Uint32 color_as_uint32;
  switch(surface->format->BytesPerPixel){ // *bytes*, not bits, per pixel
  case 1: case 2: case 4:
    color_as_uint32 = SDL_MapRGB(surface->format, r, g, b);
  } // switch
  
  /* Actually draw the pixel; the exact way of doing it depends on the
     color depth: */
  switch(surface->format->BytesPerPixel){ // *bytes*, not bits, per pixel
  case 1:
    turtle_draw_pixel_8bpp_unlocked(surface, x, y, color_as_uint32);
    break;
  case 2:
    turtle_draw_pixel_15or16bpp_unlocked(surface, x, y, color_as_uint32);
    break;
  case 3:
    turtle_draw_pixel_24bpp_unlocked(surface, x, y, r, g, b);
    break;    
  case 4:
    turtle_draw_pixel_32bpp_unlocked(surface, x, y, color_as_uint32);
    break;
  default:
    turtle_fatal("I have no idea about how to draw at %i bytes per pixel",
                 surface->format->BytesPerPixel);
  } // switch

  /* Maybe we need to unlock the SDL surface: */  
  unlock_surface_if_needed(surface);
}

static void turtle_draw_pixel_possibly_as_part_of_the_cursor(const int x, const int y, const turtle_color_t color,
                                                             const bool as_part_of_the_cursor){
  /* Do nothing if the pixel is out of the window: */
  if(! is_within_bounds(the_window, x, y))
    return;
  
  /* Draw on the cursor-less surface, unless this is part of the cursor: */
  if(! as_part_of_the_cursor)
    turtle_draw_pixel_on_surface(the_window->default_surface, x, y, color);
  
  /* And also draw on the surface with the cursor, in any case: */
  turtle_draw_pixel_on_surface(the_window->surface_with_cursor, x, y, color);
  
  /* Make the new pixel visible on the window: */
  SDL_UpdateRect(the_window->default_surface, x, y, 1, 1);
}

void turtle_draw_pixel(const int x, const int y, const turtle_color_t color){
  turtle_draw_pixel_possibly_as_part_of_the_cursor(x, y, color, false);
}

void turtle_draw_cursor_pixel(const int x, const int y, const turtle_color_t color){
  turtle_draw_pixel_possibly_as_part_of_the_cursor(x, y, color, true);
}

/* Destructively swap two variables. Note that A and B are supposed to
   be variables, not other kinds of expressions. Explicit note for some
   students who may not be expert with C: this is a *macro*, not a
   function! You *can't* do this with a function, unless you use pointers.
   C is also a call-by-value language. 
   If you don't understand this explaination and you're interested of
   course you can ask me. */
#define SWAP(A, B) \
  { const typeof(A) temporary_ = A; \
    A = B; \
    B = temporary_; }

static void turtle_draw_line_possibly_as_part_of_the_cursor(const int x1_, const int y1_,
                                                       const int x2_, const int y2_,
                                                       const turtle_color_t color,
                                                       const bool as_part_of_the_cursor){
  /* In the following I assume the width to be positive; if it's
     not, then I *make* it positive by exchanging the first point with
     the last point: */
  int x1 = x1_, y1 = y1_, x2 = x2_, y2 = y2_;
  if(x1 > x2){
    SWAP(x1, x2);
    SWAP(y1, y2);
  } // if
  
  /* We need a special case for vertical lines, as we can't compute the
     incremental ratio (we would divide by zero): */
  const int width = x2 - x1;
  const int height = y2 - y1;
  if(width != 0){ // is the line vertical?
    /* No, it's not vertical, so we can make this division: */
    const float inclination = ((float)(y2 - y1)) / ((float)(x2 - x1));
    
    /* For each x compute the corresponding y and draw the pixel: */
    float delta_x;
    for(delta_x = 0; delta_x <= width; delta_x ++){
      const int delta_y = (int)(delta_x * inclination);
      turtle_draw_pixel_possibly_as_part_of_the_cursor(x1 + delta_x, y1 + delta_y, color,
                                                       as_part_of_the_cursor);
    } // for

    /* For each y compute the corresponding x and draw the pixel: */
    const float inverse_inclination = 1.0 / inclination;
    float delta_y;
    if(height > 0){
      for(delta_y = 0; delta_y <= height; delta_y ++){
        const int delta_x = (int)(delta_y * inverse_inclination);
        turtle_draw_pixel_possibly_as_part_of_the_cursor(x1 + delta_x, y1 + delta_y, color, as_part_of_the_cursor);
      } // for
    }
    else{
      for(delta_y = height; delta_y <= 0; delta_y ++){
        const int delta_x = (int)(delta_y * inverse_inclination);
        turtle_draw_pixel_possibly_as_part_of_the_cursor(x1 + delta_x, y1 + delta_y, color, as_part_of_the_cursor);
      } // for
    }
  }
  else{
    /* Yes, the line is vertical: just draw pixel per value of y: */
    if(y1 <= y2){
      int y;
      for(y = y1; y <= y2; y ++)
        turtle_draw_pixel_possibly_as_part_of_the_cursor(x1, y, color,
                                                         as_part_of_the_cursor);
    } // inner if
    else{
      int y;
      for(y = y1; y >= y2; y --)
        turtle_draw_pixel_possibly_as_part_of_the_cursor(x1, y, color,
                                                         as_part_of_the_cursor);
    }
  } // outer else
}

void turtle_draw_line(const int x1, const int y1,
                      const int x2, const int y2,
                      const turtle_color_t color){
  turtle_draw_line_possibly_as_part_of_the_cursor(x1, y1, x2, y2, color, false);
}

void turtle_draw_cursor_line(const int x1, const int y1,
                                const int x2, const int y2,
                                const turtle_color_t color){
  turtle_draw_line_possibly_as_part_of_the_cursor(x1, y1, x2, y2, color, true);
}

void turtle_clear_window(const turtle_color_t color){
  /* Convert the color into something SDL can use: */
  Uint8 r, g, b;
  turtle_color_to_8bit_components(color, &r, &g, &b);
  const Uint32 color_as_uint32 =
    SDL_MapRGB(the_window->surface_with_cursor->format, r, g, b);
  
  /* Fill a rectangle; by the definition of this function when the
     rect is NULL the whole window is filled: */
  SDL_FillRect(the_window->surface_with_cursor, NULL, color_as_uint32);
  SDL_FillRect(the_window->default_surface, NULL, color_as_uint32);
  
  /* Show the result immediately: */
  turtle_refresh();
}

/* This macro automatically generates a global const struct containing the color
   components as a static variable (so that it's not visible to other compilation
   units), and a pointer to it.
   If you want to look at the generated C code, just run
     gcc -E graphics.c
   and look at the output. Ask me if you want to understand it but have problems;
   it's quite easy. */
#define COLOR(NAME, RED, GREEN, BLUE) \
  static const struct turtle_color turtle_ ## NAME ## _struct = { RED, GREEN, BLUE }; \
  const turtle_color_t turtle_ ## NAME = (const turtle_color_t)&turtle_ ## NAME ## _struct;

/* Call the macro once per color, to generate color definitions: */
COLOR(black,   0, 0, 0)
COLOR(white,   1, 1, 1)
COLOR(red,     1, 0, 0)
COLOR(green,   0, 1, 0)
COLOR(blue,    0, 0, 1)
COLOR(cyan,    0, 1, 1)
COLOR(magenta, 1, 0, 1)
COLOR(yellow,  1, 1, 0)

